%{

Author(s):  Elia Mulas
Revision:   1.2
Date:       23/06/2023

Copyright 2023 Micro-Photon-Devices S.r.l.

SOFTWARE PRODUCT: Hermes_Matlab 1.2

Micro-Photon-Devices (MPD) expressly disclaims any warranty for the SOFTWARE PRODUCT.
The SOFTWARE PRODUCT is provided 'As Is' without any express or implied warranty of any kind,
including but not limited to any warranties of merchantability, non-infringement, or
fitness of a particular purpose. MPD does not warrant or assume responsibility for the
accuracy or completeness of any information, text, graphics, links or other items contained
within the SOFTWARE PRODUCT. MPD further expressly disclaims any warranty or representation
to Authorized Users or to any third party.
In no event shall MPD be liable for any damages (including, without limitation, lost profits,
business interruption, or lost information) rising out of 'Authorized Users' use of or inability
to use the SOFTWARE PRODUCT, even if MPD has been advised of the possibility of such damages.
In no event will MPD be liable for loss of data or for indirect, special, incidental,
consequential (including lost profit), or other damages based in contract, tort
or otherwise. MPD shall have no liability with respect to the content of the
SOFTWARE PRODUCT or any part thereof, including but not limited to errors or omissions contained
therein, libel, infringements of rights of publicity, privacy, trademark rights, business
interruption, personal injury, loss of privacy, moral rights or the disclosure of confidential
information.

%}


classdef Hermes < handle


properties (Constant, Access=protected)
    default_lib_root_dir = 'lib/'
    lib_alias = 'Hermes_SDK';
    c_success_code = 'OK'
    mat_class_version = 1.2
   
    % Hermes-specific
    row_size = 32;           % number of columns in the spad array
    num_rows = 64;
end

properties (Access=protected)
    c_handle;
    ctor_err = false;
    enuminfo = [];
    
    % Hermes-specific
    snap_num_frames = 0;    % number of frames to be acquired in SNAP mode
    data_bits = 16;          % pixeld data depth: 8 or 16 bits
    data_is_signed = false;  % true if output data is signed
end

properties (SetAccess=private, GetAccess=public)
    % Hermes-specific
    num_counters = 1;  % number of enabled counters
    num_pixels = 2048; % total number of enabled pixels
end


methods (Access=protected)

    function load_dll(obj, lib_root_dir, additional_header)
        switch(computer('arch'))
            %case 'glnxa64'
            %    lib_path = fullfile('Linux64/', [obj.lib_alias, '.so']);
            case 'win64'
                lib_path = fullfile('Win64/', [obj.lib_alias, '.dll']);
            otherwise
                error([ errid 'wrongOS'],'Unsupported platform');
        end
        
        caller_work_dir = pwd; % save current working directory for caller i.e. user
        [class_dir, ~] = fileparts(mfilename('fullpath')); % path to directory of this this class .m file
        % move one directory in
        cd(fullfile(class_dir, lib_root_dir));

        try
            mat_header_name = [obj.lib_alias, '_MatProtos.m'];
            
            if ~exist(mat_header_name, 'file')
                % first time: generate files - they can be ported
                %             on a PC with no MATLAB-compatible C compiler
                %fprintf('Compiling %s matlab header\n', obj.lib_alias);
                c_header_path = [obj.lib_alias, '.h'];
                if  isempty(additional_header) 
                    loadlibrary(lib_path, c_header_path, 'alias', obj.lib_alias, 'mfilename', mat_header_name);
                else
                    
                    % inject develpment define in sdk header
                    header_code = fileread(c_header_path);
                    if ~contains(header_code, '#define _MPD_DEV_ 1') 
                        fid = fopen(c_header_path,'w');
                        new_header_code = sprintf('%s\n\r%s', '#define _MPD_DEV_ 1 // Automagically added via Matlab', header_code);
                        fwrite(fid, new_header_code);
                        fclose(fid);
                    end
                    
                    loadlibrary(lib_path, c_header_path, 'alias', obj.lib_alias, 'mfilename', mat_header_name, 'addheader', additional_header);
                end
            else
                %fprintf('Loading pre-compiled %s matlab header\n', obj.lib_alias);
                protos_fhndl = str2func(mat_header_name(1:end-2)); % convert .m file name (without extension) to function handle
                loadlibrary(lib_path, protos_fhndl, 'alias', obj.lib_alias);
            end
            
            protos_fhndl = str2func(mat_header_name(1:end-2)); % convert .m file name (without extension) to function handle
            [~, ~, obj.enuminfo, ~] = protos_fhndl();

        catch err
            % return to original working directory so user won't get mad
            cd(caller_work_dir);
            rethrow(err);
        end
        % return to original working directory so user won't get mad
        cd(caller_work_dir);

    end


    function unload_dll(obj)
        % overwrite still "outstanding" handle
        obj.c_handle = 0;
        unloadlibrary(obj.lib_alias)
    end


    function checkError(obj, ec)
        if ~strcmp(ec, obj.c_success_code)
            ME = MException(['Hermes:' ec], ec);
            throwAsCaller(ME);
        end
    end

    function CheckEnumArg(obj, arg_name, arg_value, sdk_enum)
        if ~ischar(arg_value) 
            err_msg = sprintf('Bad argument: %s must be a string', arg_name);
            throwAsCaller(MException(['Hermes:' 'BadEnumParameter'], err_msg));
        end
        if ~any(contains(fieldnames(sdk_enum), arg_value))
            field_names_join_cell = join(fieldnames(sdk_enum), ', ');
            fields_list = field_names_join_cell{1};
            err_msg = sprintf('Bad argument: valid values for %s: %s', arg_name, fields_list);
            throwAsCaller(MException(['Hermes:' 'BadEnumParameter'], err_msg));
        end
    end
    function CheckBooleanArg(obj, arg_name, arg_value)
        if ~islogical(arg_value) 
            err_msg = sprintf('Bad argument: %s must be a logical value', arg_name);
            throwAsCaller(MException(['Hermes:' 'BadEnumParameter'], err_msg));
        end
    end

    
    
    function inst_count = get_or_incdec_instcount(obj, inc)
        persistent INST_COUNT
        if isempty(INST_COUNT)
            INST_COUNT = 0;
        end
        if nargin == 2
            assert(inc == -1 || inc == +1)
            INST_COUNT = INST_COUNT + inc;
        end
        inst_count = INST_COUNT;
    end

    function inst_count = get_instcount(obj)
        inst_count = obj.get_or_incdec_instcount();
    end

    function incdec_instcount(obj, inc)
        obj.get_or_incdec_instcount(inc);
    end

end

 %%%% CONSTRUCTOR AND DESTRUCTOR %%%%
methods (Access=public)
    % ctor
    function obj = Hermes(mode, device_id, lib_root_dir, additional_header)
        % Constr - Constructor.
        %
        % Constr(m, Device_ID)
        %
        % It allocates a memory block to contain all the information and buffers required by the Hermes. If multiple devices are connected to the computer,
        % a unique camera ID should be provided to correctly identify the camera. The camera ID can be found in the camera documentation (9 numbers and a letter)
        % and a list of connected device is printed on the screen upon calling this function. An empty string is accepted too. In this case, the first device on the list will be connected.
        %
        % Parameters:
        %     m Camera Working mode
        %     Device_ID Unique ID to identify the connected device
        % Error codes:
        %     INVALID_OP The Hermes_H points to an occupied memory location
        %     FIRMWARE_NOT_COMPATIBLE The SDK and Firmware versions are not compatible
        %     NOT_EN_MEMORY There is not enough memory to run the camera
        %
        if obj.get_instcount() == 0
            
            if nargin <= 2
                % default to the 'public' dll
                lib_root_dir = obj.default_lib_root_dir;
                additional_header = '';
            end 
            
            obj.load_dll(lib_root_dir, additional_header);
        end
       
        if nargin < 1
            mode = 'Normal';
            device_id = '';
        elseif nargin < 2
            device_id = '';
        end
        obj.CheckEnumArg('mode', mode, obj.enuminfo.CameraMode);
        

        obj.c_handle = libpointer('s_Hermes_H');
        c_handle_ptr = libpointer('s_Hermes_HPtr', obj.c_handle);

        
        
        try
            ec = calllib(obj.lib_alias, 'HermesConstr', c_handle_ptr, mode, device_id);
            obj.checkError(ec);
            obj.incdec_instcount(+1); % increment only on success
        catch ex
            obj.ctor_err = true;
            clear hndl_ptr;
            clear obj.c_handle;
            rethrow(ex);
        end

        
        %fprintf('ctor end COUNT = %i\n', obj.get_instcount());
    end

    % dtor
    function delete(obj)
        % Destr - Destructor.
        %
        % Destr()
        %
        % It deallocates the memory block which contains all the information and buffers required by the Hermes. \b WARNING the user must call the destructor before the end of the program
        % to avoid memory leaks.
        %
        % Parameters:
        %     Hermes Hermes handle
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %
        calllib(obj.lib_alias, 'HermesDestr',  obj.c_handle);

        if ~obj.ctor_err
            obj.incdec_instcount(-1);
        end

        if obj.get_instcount() <= 0 % we are the last remaning instance, let's cleanup the lib stuff
            obj.unload_dll();
        end

        %fprintf('dtor end COUNT = %i\n', obj.get_instcount());
    end

end


methods (Access=public) %%%% Hermes-SPECIFIC FUNCTIONALITY %%%%

    function SetCameraPar(obj, Exposure, NFrames, NIntegFrames, NCounters, Force8bit, Half_array, Signed_data)
        % SetCameraPar - Print an error message.
        % 
        % SetCameraPar(Exposure, NFrames, NIntegFrames, NCounters, Force8bit, Half_array, Signed_data)
        % 
        % This function behaves differently depending on the operating mode setting. In case of Normal working mode, the exposure time is fixed to 10.40 microseconds.
        % Therefore, the parameter Exposure is not considered. Longer exposures can be obtained by summing multiple frames (i.e. by setting NIntegFrames).
        % This operating mode does not degrade the signal to noise ratio. In fact, the camera does not have any read-out noise.
        % In case of Advanced mode, all the parameters are controlled by the user which can set very long exposure times.
        % The time unit of the Exposure parameter is clock cycles i.e. the exposure time is an integer number of internal clock cycles of 10 ns period.
        % For example, the value of 10 means 100 ns exposure.
        %
        % Parameters:
        %     Exposure: Exposure time for a single frame (Hardware Integration Time - HIT). The time unit is 10 ns. Meaningful only for Advanced mode. Accepted values: 1 ... 65534
        %     NFrames: Number of frames per acquisition. Meaningful only for Snap acquisition. Accepted values: 1 ... 65534
        %     NIntegFrames: Number of integrated frames. Each output frame is the result of the sum of NIntegFrames.  Accepted values: 1 ... 65534
        %     NCounters: Number of counters per pixels to be used. Accepted values: 1 ... 3
        %     Force8bit: Force 8-bit per pixel acquisition. Counts are truncated. Meaningful only for Advanced mode.
        %     Half_array: Acquire only a 32x32 array.
        %     Signed_data: If enabled, data from counters 2 and 3 are signed data with 8-bit integer part and 1 bit sign.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     OUT_OF_BOUND Exposure, NFrames and NIntegFrames must be all greater than zero and smaller than 65535

        % HermesSetCameraPar(Hermes_H Hermes, uint16_t Exposure, uint32_t NFrames, uint16_t NIntegFrames, uint16_t NCounters, State Force8bit, State Half_array, State Signed_data);
        
        obj.CheckBooleanArg('Force8bit', Force8bit);
        obj.CheckBooleanArg('Half_array', Half_array);
        obj.CheckBooleanArg('Signed_data', Signed_data);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetCameraPar', obj.c_handle, Exposure, NFrames, NIntegFrames, NCounters, Force8bit, Half_array, Signed_data);
        obj.checkError(ec)
        
        % keep record of settings
        obj.snap_num_frames = NFrames;
        obj.num_counters = NCounters;
        obj.data_is_signed = logical(Signed_data);
        if Half_array
            obj.num_pixels = 1024;
        else
            obj.num_pixels = 2048;
        end
            
    end
    function SetCameraParSubArray(obj, Exposure, NFrames, NIntegFrames, Force8bit, Npixels)
        % SetCameraParSubArray - Set the acquisition parameters for the camera when a subarray is used.
        %
        % SetCameraParSubArray(Exposure, NFrames, NIntegFrames, Force8bit, Npixels)
        %
        % This function has to be used in alternative to the function HermesSetCameraPar() when the acquisition of a subarray is needed. Subarray acquisition has few limitations:
        % - the pixels will be readout starting from the upper-left one and moving by rows of 32 pixels toward the center of the array
        % - in live mode it is possible to acquire only 1, 2, 4, 8 full rows. If Live mode is started when the camera is set for a different number of pixels, an INVALID_OP error will be returned.
        % - in snap and continuous mode any number of pixels ranging from 1 to 256 in the upper semi-array can be acquired (in any case starting from the corner and then reading by rows,
        % i.e. if 67 pixel are required 2 rows of 32 pixels + 3 pixels from the third row will be acquired), PROVIDED that the total data acquired is an integer multiple of 1024 bytes.
        % For the example above and assuming the camera is set to 8-bit/pixel, it means that 1024 frames (or multiples) must be acquired.
        % Each frame will be Npixel*10ns + 160ns long. Acquisitions with a smaller number of frames are not possible. If an acquisition is triggered with invalid values of the parameter, an INVALID_OP error will be returned.
        % - only counter number 1 is available.
        %
        % This function behaves differently depending on the operating mode setting. In case of Normal working mode, the exposure time is fixed to Npixel*10ns + 160ns.
        % Therefore, the parameter Exposure is not considered. Longer exposures can be obtained by summing multiple frames (i.e. by setting NIntegFrames).
        % This operating mode does not degrade the signal to noise ratio. In fact, the camera does not have any read-out noise.
        % In case of Advanced mode, all the parameters are controlled by the user which can set very long exposure times.
        % The time unit of the Exposure parameter is clock cycles i.e. the exposure time is an integer number of internal clock cycles of 10 ns period.
        % For example, the value of 10 means 100 ns exposure.
        %
        % Parameters:
        %     Exposure: Exposure time for a single frame (Hardware Integration Time - HIT). The time unit is 10 ns. Meaningful only for Advanced mode. Accepted values: 1 ... 65534
        %     NFrames: Number of frames per acquisition. Meaningful only for Snap acquisition. Accepted values: 1 ... 65534
        %     NIntegFrames: Number of integrated frames. Each output frame is the result of the sum of NIntegFrames.  Accepted values: 1 ... 65534
        %     Force8bit: Force 8-bit per pixel acquisition. Counts are truncated. Meaningful only for Advanced mode.
        %     Npixels: Number of pixels to be acquired. Accepted values 1 ... 256 (see limitations above for Live mode)
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     OUT_OF_BOUND Exposure, NFrames, NIntegFrames or NPixels out of bound.

        % HermesSetCameraParSubArray(Hermes_H Hermes, uint16_t Exposure, uint32_t NFrames, uint16_t NIntegFrames, State Force8bit, uint16_t Npixels);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetCameraParSubArray', obj.c_handle, Exposure, NFrames, NIntegFrames, Force8bit, Npixels);
        obj.checkError(ec)
        
        % keep record of settings
        obj.snap_num_frames = NFrames;
        obj.num_counters = 1;
        %obj.data_is_signed = logical(Signed_data);
        obj.num_pixels = Npixels;
    end
    function SetDeadTime(obj, Val)
        % SetDeadTime - Update the dead-time setting.
        %
        % SetDeadTime(Val)
        %
        % Every time a photon is detected in a pixel, that pixel remains blind for a fix amount of time which is called dead-time. This setting is user-defined and it ranges
        % from MIN_DEAD_TIME and MAX_DEAD_TIME. Only a sub-set of this range is practically selectable: a dead-time calibration is performed during the production of the device.
        % This function will set the dead-time to the closest calibrated value to Val.
        % The default dead-time value is 50 ns.
        %
        % Parameters:
        %     Val:  New dead-time value in nanoseconds
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP Unable to change the dead-time when the live-mode is ON
        % HermesSetDeadTime(Hermes_H Hermes, uint16_t Val);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetDeadTime', obj.c_handle, Val);
        obj.checkError(ec)
    end
    function SetDeadTimeCorrection(obj, s)
        % SetDeadTimeCorrection - Enable or disable the dead-time correction.
        %
        % SetDeadTimeCorrection(s)
        %
        % The default setting is disabled.
        %
        % Parameters:
        %     s: New state for the dead-time corrector
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location

        % HermesSetDeadTimeCorrection(Hermes_H Hermes, State s);
        obj.CheckBooleanArg('s', s);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetDeadTimeCorrection', obj.c_handle, s);
        obj.checkError(ec)
    end
    function SetAdvancedMode(obj, s)
        % SetAdvancedMode - Change the operating mode.
        %
        % SetAdvancedMode(s)
        %
        % Set the operating mode to Normal or Advanced. Normal mode is the default setting. Before starting a new acquisition check that the relevant parameters for the selected modality are correctly set.
        %
        % Parameters:
        %     s: Enable or disable the advanced mode
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location

        % HermesSetAdvancedMode(Hermes_H Hermes, State s);
        obj.CheckBooleanArg('s', s);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetAdvancedMode', obj.c_handle, s);
        obj.checkError(ec)
    end
    function [] = SetBackgroundImg(obj, Img)
        % SetBackgroundImg - Load a background image to perform hardware background subtraction.
        %
        % SetBackgroundImg(Img)
        %
        % The control electronics is capable of performing real-time background subtraction. A background image is loaded into the internal camera memory.
        %
        % Parameters:
        %     Img:  Pointer to a 2048 uint16_t array containing the background image. \b WARNING The user should check the array size to avoid the corruption of the memory heap.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or Img point to an empty memory location
        %     INVALID_OP Unable to set the background image when the live-mode is ON

        % HermesSetBackgroundImg(Hermes_H Hermes, uint16_t* Img);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetBackgroundImg', obj.c_handle, Img);
        % Img
        obj.checkError(ec)
    end
    function SetBackgroundSubtraction(obj, s)
        % SetBackgroundSubtraction - Enable or disable the hardware background subtraction.
        %
        % SetBackgroundSubtraction(s)
        %
        % Parameters:
        %     s Enable or disable the background subtraction
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %

        % HermesSetBackgroundSubtraction(Hermes_H Hermes, State s);
        obj.CheckBooleanArg('s', s);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetBackgroundSubtraction', obj.c_handle, s);
        obj.checkError(ec)
    end
    function SetGateMode(obj, counter, Mode)
        % SetGateMode - Set the gate mode to continuous, coarse or pulsed (only counter 1)
        %
        % SetGateMode(counter, Mode)
        %
        % Parameters:
        %     counter Counter to which settings refer. Accepted values: 1..3
        %     Mode New gate mode
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP Only counter 1 can be set to Pulsed mode, for fast gating also counter 2 and 3 refers to HermesSetDualGate() and HermesSetTripleGate() functions.
        %

        % HermesSetGateMode(Hermes_H Hermes, uint16_t counter, GateMode Mode);
        obj.CheckEnumArg('Mode', Mode, obj.enuminfo.GateMode);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetGateMode', obj.c_handle, counter, Mode);
        obj.checkError(ec)
    end
    function SetGateValues(obj, Shift, Length)
        % SetGateValues - Change the fast Gate settings for counter 1.
        %
        % SetGateValues(Shift, Length)
        %
        % A gate signal is generated within the control electronics to select valid photons for counter 1, i.e. only photons which arrives when the Gate is ON
        % are counted. The gate signal is a 50 MHz square wave: shift and length define the phase and duty-cycle of the signal.
        %
        % Parameters:
        %     Shift: Phase shift of the gate signal in the ON state. The unit is thousandths, i.e. 10 means a delay time of 0.01 times a 20 ns periodic signal,
        %     Length: Duration of the ON gate signal. The unit is percentage.  Accepted values: 0 ... 100
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     OUT_OF_BOUND Shift or length are outside the valid values


        % HermesSetGateValues(Hermes_H Hermes, int16_t Shift, int16_t Length);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetGateValues', obj.c_handle, Shift, Length);
        obj.checkError(ec)
    end
    function SetDualGate(obj, DualGate_State, StartShift, FirstGateWidth, SecondGateWidth, Gap)
        % SetDualGate - Set parameters for DualGate mode.
        %
        % SetDualGate(DualGate_State, StartShift, FirstGateWidth, SecondGateWidth, Gap)
        %
        % In this mode two counters are used, both in gated mode. Position and width of Gate 1 and width of Gate 2 can be set by user, whereas position of Gate 2
        % is automatically set at the end of Gate 1 plus a gap selected by the user, but not smaller than 2ns. Total duration of Gate 1 and Gate 2 plus gap can not exceed 90% of the gate period of 20ns.
        %
        % Parameters:
        %     DualGate_State: Enable or disable dual-gate mode
        %     StartShift: Start delay for the first gate in thousandths of gate period (20ns). Accepted values: -500 ... +500
        %     FirstGateWidth: Duration of the ON gate 1 signal. The unit is percentage.  Accepted values: 0 ... 100
        %     SecondGateWidth: Duration of the ON gate 2 signal. The unit is percentage.  Accepted values: 0 ... 100
        %     Gap: Gap between the two gates in thousandths of nominal gate period (20ns). Accepted values: 100 ... 1000
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP This mode is not compatible with FLIM mode.
        %     OUT_OF_RANGE Parameters are out of bound. Please note that the function not only checks if the single parameters are acceptable, but also checks if the combination of parameters would result in an invalid gate


        % HermesSetDualGate(Hermes_H Hermes, State DualGate_State, int StartShift, int FirstGateWidth, int SecondGateWidth, int Gap);
        
        obj.CheckBooleanArg('DualGate_State', DualGate_State);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetDualGate', obj.c_handle, DualGate_State, StartShift, FirstGateWidth, SecondGateWidth, Gap);
        obj.checkError(ec)
    end
    function SetTripleGate(obj, TripleGate_State, StartShift, FirstGateWidth, SecondGateWidth, ThirdGateWidth, Gap1, Gap2)
        % SetTripleGate - Set parameters for TripleGate mode.
        %
        % SetTripleGate(TripleGate_State, StartShift, FirstGateWidth, SecondGateWidth, ThirdGateWidth, Gap1, Gap2)
        %
        % In this mode three counters are used, all in gated mode. The three gates cannot overlap (even over different periods), and they have to follow the order: Gate1, Gate 3, Gate2.
        % Position and width of Gate 1, and width of Gate 3 and 2 can be set by user, whereas position of Gate 3 and Gate 2 are automatically set by the function depending on the Gap1 and Gap2 values
        % specified by the user. Gap1 between Gate1 and Gate3 can be as low as 0, Gap2 between Gate3 and Gate2 must be higher than 2ns. Total duration of Gate1, Gate3, Gate2 plus Gap1 and Gap2 can not exceed 90% of the gate period of 20ns.
        %
        % Parameters:
        %     TripleGate_State: Enable or disable triple-gate mode
        %     StartShift: Start delay for the first gate in thousandths of gate period (20ns). Accepted values: -500 ... +500
        %     FirstGateWidth: Duration of the ON gate 1 signal. The unit is percentage.  Accepted values: 0 ... 100
        %     SecondGateWidth: Duration of the ON gate 3 signal. The unit is percentage.  Accepted values: 0 ... 100
        %     ThirdGateWidth: Duration of the ON gate 2 signal. The unit is percentage.  Accepted values: 0 ... 100
        %     Gap1: Gap between the gate1 and gate3 in thousandths of nominal gate period (20ns). Accepted values: 0 ... 1000
        %     Gap2: Gap between the gate3 and gate2 in thousandths of nominal gate period (20ns). Accepted values: 100 ... 1000
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP This mode is not compatible with FLIM mode.
        %     OUT_OF_RANGE Parameters are out of bound. Please note that the function not only checks if the single parameters are acceptable, but also checks if the combination of parameters would result in an invalid gate

        % HermesSetTripleGate(Hermes_H Hermes, State TripleGate_State, int StartShift, int FirstGateWidth, int SecondGateWidth, int ThirdGateWidth, int Gap1, int Gap2);
        obj.CheckBooleanArg('TripleGate_State', TripleGate_State);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetTripleGate', obj.c_handle, TripleGate_State, StartShift, FirstGateWidth, SecondGateWidth, ThirdGateWidth, Gap1, Gap2);
        obj.checkError(ec)
    end
    function SetCoarseGateValues(obj, Counter, Start, Stop)
        % SetCoarseGateValues - Change the coarse Gate settings.
        %
        % SetCoarseGateValues(Counter, Start, Stop)
        %
        % A gate signal is generated within the control electronics to select valid photons, i.e. only photons which arrives when the Gate is ON
        % are counted. The gate signal has a period equal to the hardware integration time, and the start and stop time of the ON period can be adjusted with 10ns steps.
        % Different gate settings can be applied to the 3 counters.
        %
        % Parameters:
        %     Counter: Counter to which settings refer. Accepted values: 1..3
        %     Start: Starting position of the ON period. Can range from 0 to (HIT - 6), where units is 10ns and HIT is the Hardware Integration Time set with HermesSetCameraPar().
        %     Stop: Stop position of the ON period. Can range from (Start+1) to (HIT - 5), where units is 10ns and HIT is the Hardware Integration Time set with HermesSetCameraPar().
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     OUT_OF_BOUND Start or Stop are outside the valid values

        % HermesSetCoarseGateValues(Hermes_H Hermes, uint16_t Counter, uint16_t Start, uint16_t Stop);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetCoarseGateValues', obj.c_handle, Counter, Start, Stop);
        obj.checkError(ec)
    end
    function SetTriggerOutState(obj, Mode)
        % SetTriggerOutState - Select the output signal.
        %
        % SetTriggerOutState(Mode)
        %
        % Parameters:
        %     Mode New trigger mode
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %

        % HermesSetTriggerOutState(Hermes_H Hermes, TriggerMode Mode);
        obj.CheckEnumArg('Mode', Mode, obj.enuminfo.TriggerMode);
        
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetTriggerOutState', obj.c_handle, Mode);
        obj.checkError(ec)
    end
    function SetSyncInState(obj, s, frames)
        % SetSyncInState - Set the sync-in state.
        %
        % SetSyncInState(s, frames)
        %
        % Set the camera to wait for an input trigger signal before starting an acquisition.
        %
        % Parameters:
        %     s: Enable or disable the synchronization input
        %     frames: If the synchronization input is enabled, this is the number of frames that are acquired for each pulse (0 means that the acquisition will wait only the first pulse and then continue to the end with no further pauses). Accepted values: 0 ... 100.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location

        % HermesSetSyncInState(Hermes_H Hermes, State s, int frames);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetSyncInState', obj.c_handle, s, frames);
        obj.checkError(ec)
    end
    function LiveSetModeON(obj)
        % LiveSetModeON - Turn on the Live mode.
        %
        % LiveSetModeON()
        %
        % The camera is set in the Live mode, i.e. it continuously acquires images (free-running mode). The frames which are not transferred to the computer are discarded.
        % Therefore, the time-laps between two frames is not constant and it will depend on the transfer speed between the host computer and the camera.
        % This mode is very useful to adjust optical components or to align the camera position. When the camera is in Live mode, no acquisition of images by HermesSnapAcquire()
        % or HermesContAcqToFileGetMemory() can be performed.
        %
        % Parameters:
        %     None.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP The live mode has been already started

        % HermesLiveSetModeON(Hermes_H Hermes);
        [ec, ~] = calllib(obj.lib_alias, 'HermesLiveSetModeON', obj.c_handle);
        obj.checkError(ec)
    end
    function LiveSetModeOFF(obj)
        % LiveSetModeOFF - Turn off the Live mode.
        %
        % LiveSetModeOFF()
        %
        % Parameters:
        %     None.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP The live mode is already inactive
        %

        % HermesLiveSetModeOFF(Hermes_H Hermes);
        [ec, ~] = calllib(obj.lib_alias, 'HermesLiveSetModeOFF', obj.c_handle);
        obj.checkError(ec)
    end
    function FLIM_frame_time = SetFlimPar(obj, FLIM_steps, FLIM_shift, FLIM_start, Length)
        % SetFlimPar - Set FLIM parameters.
        %
        % SetFlimPar(FLIM_steps, FLIM_shift, FLIM_start, Length, FLIM_frame_time)
        %
        % The camera can perform automatic time-gated FLIM measurements employing the embedded gate generator. Call this function to setup the FLIM acquisition parameters. Each "FLIM acquisition" is composed by FLIM_steps frames,
        % each one consisting of an acquisition with Exposure and NIntegFrames as set with HermesSetCameraPar(). The total time required to perform each FLIM acquisition is passed back to the caller through the referenced FLIM_frame_time variable.
        %
        % Parameters:
        %     FLIM_steps: Number of gate delay steps to be performed. Accepted values: 1 ... 1000
        %     FLIM_shift: Delay shift between steps in thousandths of gate period (20ns). Accepted values: 1 ... 1000
        %     FLIM_start: Start delay for FLIM sequence in thousandths of gate period (20ns). Accepted values: -500 ... +500
        %     Length: Duration of the ON gate signal. The unit is percentage.  Accepted values: 0 ... 100
        %
        % Returns:
        %     Total time required to perform each FLIM acquisition in multiples of 10ns.
        %
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     OUT_OF_BOUND Parameters are out of bound. Please note that the function not only checks if the single parameters are acceptable, but also checks if the combination of parameters would result in an invalid gate


        % HermesSetFlimPar(Hermes_H Hermes, uint16_t FLIM_steps, uint16_t FLIM_shift, int16_t FLIM_start, uint16_t Length, int* FLIM_frame_time);
        [ec, ~, FLIM_frame_time] = calllib(obj.lib_alias, 'HermesSetFlimPar', obj.c_handle, FLIM_steps, FLIM_shift, FLIM_start, Length, int32(0));
        % FLIM_frame_time
        obj.checkError(ec)
    end
    function SetFlimState(obj, FLIM_State)
        % SetFlimState - Enable or disable FLIM mode. FLIM mode automatically set the number of used counters to 1. FLIM mode cannot be enabled if Exposure time is set to a value lower than 1040.
        %
        % SetFlimState(FLIM_State)
        %
        % Parameters:
        %     FLIM_State Enable or disable the FLIM mode
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location.
        %     INVALID_OP Exposure time is lower than 1040.
        %

        % HermesSetFlimState(Hermes_H Hermes, State FLIM_State);
        obj.CheckBooleanArg('FLIM_State', FLIM_State);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetFlimState', obj.c_handle, FLIM_State);
        obj.checkError(ec)
    end
    function ApplySettings(obj)
        % ApplySettings - Apply settings to the camera.
        %
        % ApplySettings()
        %
        % This function must be called after any Set function, except HermesLiveSetModeON() and HermesLiveSetModeOFF(), in order to apply the settings to the camera.
        % If several Set functions need to be called, there is no need to call this function after each Set function. A single call to this function at the end is enough to apply all the settings.
        %
        % Parameters:
        %     None.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location


        % HermesApplySettings(Hermes_H Hermes);
        [ec, ~] = calllib(obj.lib_alias, 'HermesApplySettings', obj.c_handle);
        obj.checkError(ec)
        
        if obj.Is16Bit()
            obj.data_bits = 16;
        else
            obj.data_bits = 8;
        end
    end
    function frames = LiveGetImg(obj)
        % LiveGetImg - Get a Live image for each active counter.
        %
        % LiveGetImg(obj)
        %
        % Parameters:
        %     None
        %
        % Returns:
        %     Live image for each active counter.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or Img point to an empty memory location
        %     INVALID_OP The live-mode has not been started yet
        

        % HermesLiveGetImg(Hermes_H Hermes, uint16_t* Img);
        
        data = zeros(obj.row_size * obj.num_rows * obj.num_counters, 1, 'uint16');
        [ec, ~, data] = calllib(obj.lib_alias, 'HermesLiveGetImg', obj.c_handle, data);
        obj.checkError(ec)
        
        frames = obj.BufferToFrames(data, obj.num_pixels , obj.num_counters);
        % remove frame index dimension
        frames = permute(frames, [1,2,4,3]); % shift if as last dimension
        frames = frames(:,:,:,1); % index it out
    end
    function SnapPrepare(obj)
        % SnapPrepare - Prepare the camera to the acquisition of a snap.
        %
        % SnapPrepare()
        %
        % This command configures the camera to acquire a snap of NFrames images, as set by the HermesSetCameraPar() function. In FLIM mode NFrames "FLIM acquisitions" of a FLIM sequence will be acquired.
        % If an External Sync is required, the camera will wait for a pulse on the Sync input before acquiring the images and saving them to the
        % internal memory, otherwise they are acquired and saved immediately. Once acquired, snap must then be transferred to the PC using the HermesSnapAcquire() function.
        %
        % Parameters:
        %     None.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP Unable to acquire images when the live mode is ON. Use instead HermesLiveGetImg().
        %     INVALID_OP When the background subtraction, dead-time correction or normal acquisition mode are enabled,

        % HermesSnapPrepare(Hermes_H Hermes);
        if obj.num_pixels < 1024 % subarray mode
            data_len = 2 * obj.snap_num_frames * obj.data_bits/8 * obj.num_pixels;
        else
            data_len = obj.snap_num_frames * obj.data_bits/8 * obj.num_pixels * obj.num_counters;
        end
        
        if rem(data_len, 1024) == 0
            [ec, ~] = calllib(obj.lib_alias, 'HermesSnapPrepare', obj.c_handle);
            obj.checkError(ec)
        else
            error('Hermes:invalidSettings', 'Invalid settings: requested data length (%.0f bytes) is not divisible by 1024', data_len);
        end
    end
    function SnapAcquire(obj)
        % SnapAcquire - Get a selected number of images.
        %
        % SnapAcquire()
        %
        % Acquire a set of images according to the parameters defined by HermesSetCameraPar(). In FLIM mode NFrames "FLIM acquisitions" will be acquired.
        % This command works only when HermesSnapPrepare() has already been called.
        % This function will not exit until the required number of images has been downloaded. For this reason,
        % if the camera is configured for waiting and External Sync, before calling this function it could be useful to poll the camera for the trigger state,
        % using the HermesIsTriggered() function.

        % Parameters:
        %     None.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP Unable to acquire images when the live mode is ON. Use instead HermesLiveGetImg().
        %     INVALID_OP When the background subtraction, dead-time correction or normal acquisition mode are enabled,
        %

        % HermesSnapAcquire(Hermes_H Hermes);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSnapAcquire', obj.c_handle);
        obj.checkError(ec)
    end
    function frames = SnapGetImageBuffer(obj)
        % SnapGetImageBuffer - Get the pointer to the image buffer in which snap acquisition is stored. Refer to Figure 6 of the User Manual for pixels readout order and position in the image. \b WARNING User must pay attention not to exceed the dimension of the buffer (2*1024*65534 bytes) when accessing it.
        %
        % SnapGetImageBuffer(buffer, DataDepth)
        %
        %
        %
        % Parameters:
        %     None
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or BUFFER_H point to an empty memory location
        
        buf = uint8([]);
        % HermesSnapGetImageBuffer(Hermes_H Hermes, BUFFER_H* buffer, int* DataDepth);
        [ec, ~, buf, DataDepth] = calllib(obj.lib_alias, 'HermesSnapGetImageBuffer', obj.c_handle, buf, int32(0));
        % buffer, DataDepth
        obj.checkError(ec)

        size = obj.snap_num_frames * obj.data_bits/8 * obj.num_pixels * obj.num_counters;
        
        if size > 2*1024*65534
            error('Hermes:tooMuchData', 'Expected data is larger than maximum available buffer size.');
        end
        
        if size > 0
            buf.reshape(size,1);
            data = buf.Value;
        else
            data = [];
        end
            
        if obj.data_bits == 16
            data = typecast(data, 'uint16');
        end
        
        frames = Hermes.BufferToFrames(data, obj.num_pixels, obj.num_counters);
        
    end
    function Img = SnapGetImgPosition(obj, Position, counter)
        % SnapGetImgPosition - Export an acquired image.
        %
        % SnapGetImgPosition(Img, Position, counter)
        %
        % Once a set of images have been acquired by HermesSnapAcquire(), a single image can be exported from the SDK image buffer and saved
        % in the memory (Img array).
        %
        % Parameters:
        %     Img: Pointer to the output image array. The size of the array must be at least 4kB.
        %     Position: Index of the image to save.  Accepted values: 1 ... Number of acquired images
        %     counter: Number of the desired counter. Accepted values: 1 ... Number of used counters
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or Img point to an empty memory location
        %     OUT_OF_BOUND Parameters are out of bound.


        data = zeros(obj.row_size * obj.num_rows, 1, 'uint16'); % allocate the maximum required memory space
        % HermesSnapGetImgPosition(Hermes_H Hermes, uint16_t* Img, uint32_t Position, uint16_t counter);
        [ec, ~, data] = calllib(obj.lib_alias, 'HermesSnapGetImgPosition', obj.c_handle, data, Position, counter);
        % Img
        obj.checkError(ec)
        frames = obj.BufferToFrames(data, obj.num_pixels , 1); % expect data of just one counter!
        
        Img = frames(:,:,1,1);
        obj.checkError(ec)
    end
    function [] = ContAcqToFileStart(obj, filename)
        % ContAcqToFileStart - Put the camera in "continuous acquisition" mode. Compatible with FLIM mode. If the camera was set to wait for an external sync, the acquisition will start as soon as a pulse is detected on the Sync input, otherwise it
        %
        % ContAcqToFileStart(filename)
        %
        % will start immediately.  The output file name must be provided when calling this function. Data are stored in the camera internal memory and must be downloaded calling the HermesContAcqToFileGetMemory() function
        % as soon as possible, in order to avoid data loss.
        %
        % Parameters:
        %     filename: Full path of the output file. The string length must not exceed 1024 characters.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or BUFFER_H point to an empty memory location
        %     UNABLE_CREATE_FILE It was not possible to create the output file.

        % HermesContAcqToFileStart(Hermes_H Hermes, char* filename);
        [ec, ~] = calllib(obj.lib_alias, 'HermesContAcqToFileStart', obj.c_handle, filename);
        % filename
        obj.checkError(ec)
    end
    function total_bytes = ContAcqToFileGetMemory(obj)
        % ContAcqToFileGetMemory - Dump the camera memory to the PC and save data to the file specified with the HermesContAcqToFileStart() function in Hermes file format (for details on the format see function HermesSaveImgDisk).
        % This function must be repeatedly called, as fast as possible, in order to free the camera
        % internal memory and keep the acquisition going. If the internal camera memory get full during acquisition an error is generated. \b WARNING The camera can generate data with very high throughput,
        % up to about 205MB/s. Be sure to have enough disk space for your measurement.
        %
        % ContAcqToFileGetMemory(total_bytes)
        %
        % Parameters:
        %     None.
        % Returns:
        %     Total number of bytes read.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or BUFFER_H point to an empty memory location
        %     UNABLE_CREATE_FILE It was not possible to access the output file.
        %     INVALID_OP Continuous acquisition was not yet started. Use Hermes_HermesContAcqToFileStart() before calling this function.
        %     COMMUNICATION_ERROR Communication error during data download.
        %     Hermes_MEMORY_FULL Camera internal memory got full during data download. Data loss occurred. Reduce frame-rate or optimize your software to reduce dead-time between subsequent calling of the function.


        % HermesContAcqToFileGetMemory(Hermes_H Hermes, double* total_bytes);
        [ec, ~, total_bytes] = calllib(obj.lib_alias, 'HermesContAcqToFileGetMemory', obj.c_handle, double(0));
        % total_bytes
        obj.checkError(ec)
    end
    function ContAcqToFileStop(obj)
        % ContAcqToFileStop - Stop the continuous acquisition of data and close the output file. This function must be called at the end of the continuous acquisition, in order to properly close the file. \b WARNING If not called,
        %
        % ContAcqToFileStop()
        %
        % the output file may be unreadable, and camera may have unexpected behavior if other functions are called beforehand.
        %
        % Parameters:
        %     None.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or BUFFER_H point to an empty memory location
        %     UNABLE_CREATE_FILE It was not possible to access the output file.

        % HermesContAcqToFileStop(Hermes_H Hermes);
        [ec, ~] = calllib(obj.lib_alias, 'HermesContAcqToFileStop', obj.c_handle);
        obj.checkError(ec)
    end
    function ContAcqToMemoryStart(obj)
        % ContAcqToMemoryStart - Put the camera in "continuous acquisition" mode. Compatible with FLIM mode. If the camera was set to wait for an external sync, the acquisition will start as soon as a pulse is detected on the Sync input, otherwise it
        %
        % ContAcqToMemoryStart()
        %
        % will start immediately. Data are stored in the camera internal memory and must be downloaded calling the HermesContAcqToMemoryGetBuffer() function as soon as possible, in order to avoid data loss.
        %
        % Parameters:
        %     None.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or BUFFER_H point to an empty memory location

        % HermesContAcqToMemoryStart(Hermes_H Hermes);
        [ec, ~] = calllib(obj.lib_alias, 'HermesContAcqToMemoryStart', obj.c_handle);
        obj.checkError(ec)
    end
    function data = ContAcqToMemoryGetBuffer(obj)
        % ContAcqToMemoryGetBuffer - Dump the camera memory to the PC and pass the pointer to the image buffer in which acquisition is stored. This function must be repeatedly called, as fast as possible, in order to free the camera internal memory and keep the acquisition going.
        %
        % ContAcqToMemoryGetBuffer(total_bytes, buffer)
        %
        % If the internal camera memory get full during acquisition an error is generated. Refer to Figure 6 of the User Manual for pixels readout order and position in the image. \b WARNING User must pay attention not to exceed the dimension of the buffer when accessing it. Refer to the value saved into the parameter total_bytes.
        %
        % Parameters:
        %     None.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or BUFFER_H point to an empty memory location
        %     INVALID_OP Continues acquisition was not yet started. Use HermesContAcqToMemoryStart() before calling this function.
        %     COMMUNICATION_ERROR Communication error during data download.
        %     Hermes_MEMORY_FULL Camera internal memory got full during data download. Data loss occurred. Reduce frame-rate or optimize your software to reduce deadtime between subsequent calling of the function.



        buf = libpointer('uint8Ptr'); % BUFFER_H* is uint8**
        % HermesContAcqToMemoryGetBuffer(Hermes_H Hermes, double* total_bytes, BUFFER_H* buffer);
        [ec, ~, size] = calllib(obj.lib_alias, 'HermesContAcqToMemoryGetBuffer', obj.c_handle, double(0), buf);
        % total_bytes, buffer
        obj.checkError(ec)

        if size > 0
            buf.reshape(size,1);
            data = buf.Value;
        else
            data = [];
        end

        if obj.data_bits == 16
            data = typecast(data, 'uint16');
        end

    end
    
    function ContAcqToMemoryStop(obj)
        % ContAcqToMemoryStop - Stop the continuous acquisition of data. This function must be called at the end of the continuous acquisition. \b WARNING If not called, camera may have unexpected behavior if other functions are called.
        %
        % ContAcqToMemoryStop()
        %
        %
        %
        % Parameters:
        %     None.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H point to an empty memory location


        % HermesContAcqToMemoryStop(Hermes_H Hermes);
        [ec, ~] = calllib(obj.lib_alias, 'HermesContAcqToMemoryStop', obj.c_handle);
        obj.checkError(ec)
    end
    function ReturnVal = GetDeadTime(obj, Val)
        % GetDeadTime - Get the calibrated dead-time value.
        %
        % GetDeadTime(Val, ReturnVal)
        %
        % This function provides the closest calibrated dead-time value to Val.
        %
        % Parameters:
        %     Val: Desired dead-time value in ns. No error is generated when the value is above MAX_DEAD_TIME.
        %
        % Returns:
        %     Closest dead-time value possible.
        %
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or ReturnVal point to an empty memory location

        % HermesGetDeadTime(Hermes_H Hermes, uint16_t Val, uint16_t* ReturnVal);
        [ec, ~, ReturnVal] = calllib(obj.lib_alias, 'HermesGetDeadTime', obj.c_handle, Val, uint16(0));
        % ReturnVal
        obj.checkError(ec)
    end
    function ReturnVal = GetGateWidth(obj, counter, Val)
        % GetGateWidth - Get the calibrated gate width value.
        %
        % GetGateWidth(counter, Val, ReturnVal)
        %
        % This function provides the closest calibrated gate-width value to Val.
        %
        % Parameters:
        %     counter: Counter for which the gate width is requested. Accepted values: 1..3
        %     Val: Desired gate-width value in percentage of 20ns. No error is generated when the value is out of range, instead the real boundaries are forced on ReturnVal.
        %
        % Returns:
        %     Closest gate-width value possible.
        %
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or ReturnVal point to an empty memory location
        %

        % HermesGetGateWidth(Hermes_H Hermes, uint16_t counter, int16_t Val, double* ReturnVal);
        [ec, ~, ReturnVal] = calllib(obj.lib_alias, 'HermesGetGateWidth', obj.c_handle, counter, Val, double(0));
        % ReturnVal
        obj.checkError(ec)
    end
    function ReturnVal = GetGateShift(obj, counter, Val)
        % GetGateShift - Get the calibrated gate shift value.
        %
        % GetGateShift(counter, Val, ReturnVal)
        %
        % This function provides the closest calibrated gate shift value to Val.
        %
        % Parameters:
        %     counter: Counter for which the gate shift is requested. Accepted values: 1..3
        %     Val: Desired gate shift value in thousandths of 20ns. No error is generated when the value out of range, instead the real boundaries are forced on ReturnVal.
        % Returns:
        %     Closest gate-shift value possible.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or ReturnVal point to an empty memory location

        %

        % HermesGetGateShift(Hermes_H Hermes, uint16_t counter, int16_t Val, int16_t* ReturnVal);
        [ec, ~, ReturnVal] = calllib(obj.lib_alias, 'HermesGetGateShift', obj.c_handle, counter, Val, int16(0));
        % ReturnVal
        obj.checkError(ec)
    end
    function is16bit = Is16Bit(obj)
        % Is16Bit - Get the actual bit depth of acquired data.
        %
        % Is16Bit(is16bit)
        %
        % Data from the camera will be 16-bit per pixel, if NFramesInteg > 1, or DTC is enabled, or background subtraction is enabled, or 8-bit per pixel otherwise.
        % This function provides actual bit depth with the current settings.
        %
        % Parameters:
        %     None.
        %
        % Returns:
        %     Returns True if the camera is setp to provide 16-bit data, False otherwise.
        %
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or is16bit pointers point to an empty memory location

        % HermesIs16Bit(Hermes_H Hermes, short* is16bit);
        [ec, ~, is16bit_uint] = calllib(obj.lib_alias, 'HermesIs16Bit', obj.c_handle, int16(0));
        is16bit = logical(is16bit_uint);
        % is16bit
        obj.checkError(ec)
    end
    function isTriggered = IsTriggered(obj)
        % IsTriggered - Poll the camera for external trigger status.
        %
        % IsTriggered(isTriggered)
        %
        % Poll the camera in order to know if an external sync pulse was detected. The result is meaningful only if the camera was previously set to wait for an external sync.
        %
        % Parameters:
        %     None.
        % Returns:
        %     Actual status. The value is 0 if no sync pulse was detected so far, 1 otherwise.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H or is Triggered pointers point to an empty memory location

        % HermesIsTriggered(Hermes_H Hermes, short* isTriggered);
        [ec, ~, isTriggered_uint] = calllib(obj.lib_alias, 'HermesIsTriggered', obj.c_handle, int16(0));
        isTriggered = logical(isTriggered_uint);
        % isTriggered
        obj.checkError(ec)
    end
    function [Firmware_Version, Software_Version, Custom_version] = GetVersion(obj)
        % GetVersion - Get the SDK and camera firmware version
        %
        % GetVersion(Firmware_Version, Software_Version, Custom_version)
        %
        %
        %
        % Parameters:
        %     None.
        % Returns:
        %     Firmware_Version: Version of the camera firmare in the format x.xx.
        %     Software_Version: Version of the SDK in the format x.xx.
        %     Custom_version: Customization version of the firmware and SDK. For standard model "A" is returned.
        % Error codes:
        %     NULL_POINTER The provided handle or pointers point to an empty memory location


        % HermesGetVersion(Hermes_H Hermes, double* Firmware_Version, double* Software_Version, char* Custom_version);
        [ec, ~, Firmware_Version, Software_Version, Custom_version] = calllib(obj.lib_alias, 'HermesGetVersion', obj.c_handle, 0, 0, ' ');
        % Firmware_Version, Software_Version, Custom_version
        obj.checkError(ec)
    end
    function [Camera_ID, Camera_serial] = GetSerial(obj)
        % GetSerial - Get the camera serial number and ID
        %
        % GetSerial(Camera_ID, Camera_serial)
        %
        %
        %
        % Parameters:
        %     None.
        % Returns:
        %     Camera_ID: Unique camera ID. A string of at least 11 character is required as parameter.
        %     Camera_serial: Hermes camera serial number. A string of at least 33 character is required as parameter.
        % Error codes:
        %     NULL_POINTER The provided handle or pointers point to an empty memory location.

        % HermesGetSerial(Hermes_H Hermes, char* Camera_ID, char* Camera_serial);
        [ec, ~, Camera_ID, Camera_serial] = calllib(obj.lib_alias, 'HermesGetSerial', obj.c_handle, Camera_ID, Camera_serial);
        % Camera_ID, Camera_serial
        obj.checkError(ec)
    end
    function [Camera_serial, Firmware_Version, Software_Version, Firmware_Custom_Version, Software_Custom_Version] = DeviceInfo(obj, Device_ID)
        % DeviceInfo - Get device info.
        %
        % DeviceInfo(Camera_serial, Firmware_Version, Software_Version, Firmware_Custom_Version, Software_Custom_Version)
        %
        % It gets device serial number, Unique ID, version and SDK version, without constructing an Hermes object. Useful when constructor fails due to incompatible firmware and SDK versions or for powering issues.
        %
        % Parameters:
        %     Device_ID Hermes camera Unique ID. A string of at least 11 character is required as parameter. If multiple devices are connected to the computer, a unique camera ID should be provided to correctly identify the camera. The camera ID can be found in the camera documentation.
        %     Camera_serial Hermes camera serial number. A string of at least 33 character is required as parameter. This parameter is referenced.
        %     Firmware_Version Version of the camera firmare in the format x.xx. This parameter is referenced.
        %     Software_Version Version of the SDK in the format x.xx. This parameter is referenced.
        %     Firmware_Custom_Version Customization version of the firmware. For standard model "A" is returned. This parameter is referenced.
        %     Software_Custom_Version Customization version of the software. For standard model "A" is returned. This parameter is referenced.
        % Error codes:
        %     None.

        % HermesDeviceInfo(char* Device_ID, char* Camera_serial, double* Firmware_Version, double* Software_Version, char* Firmware_Custom_Version, char* Software_Custom_Version);
        [ec, ~, Camera_serial, Firmware_Version, Software_Version, Firmware_Custom_Version, Software_Custom_Version] = calllib(obj.lib_alias, 'HermesDeviceInfo', Device_ID, blanks(33), 0, 0, char([1 0]), char([1 0]));
        % Device_ID, Camera_serial, Firmware_Version, Software_Version, Firmware_Custom_Version, Software_Custom_Version
        obj.checkError(ec)
    end
    function [] = SaveImgDisk(obj, Start_Img, End_Img, filename, mode)
        % SaveImgDisk - Save the selected images on the hard disk.
        %
        % SaveImgDisk(Start_Img, End_Img, filename, mode)
        %
        % This function saves the acquired images on the hard disk. The output file format can be either a multipage TIFF with embedded acquisition metadata according to the OME-TIFF format or the proprietary Hermes format. For FLIM measurements, use the HermesSaveFlimDisk() function.
        % If TIFF format is selected, the desired images will be saved in a file for each enabled counter. If Hermes format is selected a single Hermes file will be created for all the counters.
        % OME-TIFF file could be opened with any image reader compatible with TIFF file, since metadata are saved into the Image Description tag in XML format.
        % In order to decode OME-TIFF metadata, it is possible to use a free OME-TIFF reader, such as OMERO or the Bio-Formats plugin for ImageJ. For more details see the OME-TIFF web site:
        % http://www.openmicroscopy.org/site/support/ome-model/ome-tiff/. If subarray acquisition is enabled and the number of pixels is not an integer multiple of 32,
        % the TIFF files will have as much rows of 32 pixels as needed to accommodate all pixels, and the missing pixels will be put to 0, e.g.  if 67 pixels are acquired, the TIFF image will be 32x3, with the last 29 pixels of the 3rd row set to 0.
        % Hermes file are binary files composed by a header with acquisition metadata followed by raw image data, containing the 8/16 bit (integer) or 64 bit (double precision) pixel values in row-major order (refer to Figure 7 of the User Manual for pixels position and order). The byte order is little-endian for the 16 or 64 bit images.
        % In case more counters are used, data are interlaced, i.e. the sequence of frames is the following:
        % 1st frame of 1st counter, 1st frame of 2nd counter, 1st frame of 3rd counter, 2nd frame of 1st counter, etc.
        % The header is composed by a signature of 8 byte (0x4d5044ff04000000, starting with 4d on byte 0), and a metadata section of 1024 byte, as follows (multibyte fields are little-endian):
        %
        % | Byte offset   | Number of bytes   | Description                                    |
        % | --------------|-------------------|-----------------------------------------------|
        % | 0             | 10                   | Unique camera ID (string)                        |
        % | 10            | 32                | Hermes serial number (string)                    |
        % | 42            | 2                    | Firmware version (x.xx saved as xxx)            |
        % | 44            | 1                    | Firmware custom version (standard = 0)        |
        % | 45            | 20                | Acquisition data&time (string)                |
        % | 65            | 35                | Unused                                        |
        % | 100           | 1                 | Number of rows                                |
        % | 101           | 1                 | Number of columns                                |
        % | 102           | 1                 | Bit per pixel                                    |
        % | 103           | 1                 | Counters in use                                |
        % | 104           | 2                 | Hardware integration time    (multiples of 10ns)    |
        % | 106           | 2                 | Summed frames                                     |
        % | 108           | 1                 | Dead time correction enabled                    |
        % | 109           | 1                 | Internal gate duty-cycle for counter1 (0-100%)|
        % | 110           | 2                 | Hold-off time (ns)                            |
        % | 112            | 1                 | Background subtraction enabled                |
        % | 113           | 1                 | Data for counters 1 and 2 are signed            |
        % | 114            | 4                    | Number of frames in the file                    |
        % | 118            | 1                    | Image is averaged                                |
        % | 119            | 1                    | Counter which is averaged                        |
        % | 120            | 2                    | Number of averaged images                        |
        % | 122           | 1                 | Internal gate duty-cycle for counter2 (0-100%)|
        % | 123           | 1                 | Internal gate duty-cycle for counter3 (0-100%)|
        % | 124            | 2                    | Frames per sync-in pulse                        |
        % | 126            | 2                    | Number of pixels                                |
        % | 128            | 72                | Unused                                        |
        % | 200           | 1                 | FLIM enabled                                    |
        % | 201           | 2                 | FLIM shift (thousandths of gate period)        |
        % | 203           | 2                 | FLIM steps                                    |
        % | 205           | 4                 | FLIM frame length (multiples of 10ns)            |
        % | 209           | 2                 | FLIM bin width (fs)                           |
        % | 211            | 9                    | Unused                                        |
        % | 220            | 1                    | Multi gate mode: 2 = dual, 3 = triple            |
        % | 221            | 2                    | Multi gate mode: start position (-500 - +500)    |
        % | 223            | 1                    | Multi gate mode: first gate width (0-100%)    |
        % | 224            | 1                    | Multi gate mode: second gate width (0-100%)     |
        % | 225            | 1                    | Multi gate mode: third gate width (0-100%)     |
        % | 226            | 2                    | Multi gate mode: gap1 (0-800)                    |
        % | 228            | 2                    | Multi gate mode: gap2 (0-800)                    |
        % | 230            | 2                    | Multi gate mode: calibrated bin-width in fs    |
        % | 232            | 1                    | Coarse gate 1 enabled                            |
        % | 233            | 2                    | Coarse gate 1 start                            |
        % | 235            | 2                    | Coarse gate 1 stop                            |
        % | 237            | 1                    | Coarse gate 2 enabled                            |
        % | 238            | 2                    | Coarse gate 2 start                            |
        % | 240            | 2                    | Coarse gate 2 stop                            |
        % | 242            | 1                    | Coarse gate 3 enabled                            |
        % | 243            | 2                    | Coarse gate 3 start                            |
        % | 245            | 2                    | Coarse gate 3 stop                            |
        % | 247            | 53                | Unused                                        |
        % | 300           | 1                 | PDE measurement                                |
        % | 301           | 2                 | Start wavelength (nm)                            |
        % | 303           | 2                 | Stop wavelength (nm)                            |
        % | 305           | 2                 | Step (nm)                                        |
        % | 307           | 717               | unused                                        |
        %
        % Hermes file can be read using the provided ImageJ/Fiji plugin.
        %
        % Parameters:
        %     Start_Img: Index of the first image to save.  Accepted values: 1 ... Number of acquired images
        %     End_Img: Index of the last image to save.  Accepted values: Start_Img ... Number of acquired images
        %     filename: Full path of the output file.
        %     mode: File format of the output images
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP No images were acquired or the selected range of images is not valid
        %     UNABLE_CREATE_FILE Unable to create the output file

        % HermesSaveImgDisk(Hermes_H Hermes, uint32_t Start_Img, uint32_t End_Img, char* filename, OutFileFormat mode);
        obj.CheckEnumArg('mode', mode, obj.enuminfo.OutFileFormat);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSaveImgDisk', obj.c_handle, Start_Img, End_Img, filename, mode);
        % filename
        obj.checkError(ec)
    end
    function SaveAveragedImgDisk(obj, counter, filename, mode, isDouble)
        % SaveAveragedImgDisk - Save the selected images on the hard disk.
        %
        % SaveAveragedImgDisk(counter, filename, mode, isDouble)
        %
        % This function saves the average of the images acquired by a specified counter on the hard disk. File format can be proprietary Hermes or TIFF, as explained in HermesSaveImgDisk() function.
        %
        % Parameters:
        %     counter Number of the counter to be saved. Accepted values: 1..3
        %     filename Full path of the output file. Value is referenced.
        %     mode File format of the output images
        %     isDouble Number format. 0 for Uint16, 1 for Double
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP No images were acquired or the selected range of images is not valid
        %     UNABLE_CREATE_FILE Unable to create the output file
        %

        % HermesSaveAveragedImgDisk(Hermes_H Hermes, uint16_t counter, char* filename, OutFileFormat mode, short isDouble);
        obj.CheckEnumArg('mode', mode, obj.enuminfo.OutFileFormat);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSaveAveragedImgDisk', obj.c_handle, counter, filename, mode, isDouble);
        % filename
        obj.checkError(ec)
    end
    function [] = SaveFlimDisk(obj, filename, mode)
        % SaveFlimDisk - Save the FLIM acquisition on the hard disk.
        %
        % SaveFlimDisk(filename, mode)
        %
        % This function saves the acquired FLIM images on the hard disk. The output file format can be either a multipage TIFF with embedded acquisition metadata according to the OME-TIFF format or the proprietary Hermes format. For standard measurements, use the Hermes_Save_Img__Disk() function.
        % For both formats, image data is composed by a set of images following a "FLIM first, time second scheme", i.e. with the following frame sequence:
        % 1st gate shift of 1st FLIM measurement, 2nd gate shift of 1st FLIM measurement,...,nth gate shift of 1st FLIM measurement,1st gate shift of 2nd FLIM measurement, 2nd gate shift of 2nd FLIM measurement,...,nth gate shift of 2nd FLIM measurement,etc.
        % OME-TIFF file could be opened with any image reader compatible with TIFF file, since metadata are saved into the Image Description tag in XML format.
        % In order to decode OME-TIFF metadata, it is possible to use a free OME-TIFF reader, such as OMERO or the Bio-Formats plugin for ImageJ. For more details see the OME-TIFF web site:
        % http://www.openmicroscopy.org/site/support/ome-model/ome-tiff/. OME-TIFF metadata include the ModuloAlongT tag, which allows the processing of FLIM data with dedicated FLIM software such as FLIMfit (see http://www.openmicroscopy.org/site/products/partner/flimfit).
        % Hermes file are binary files composed by a header with acquisition metadata followed by raw image data, containing the 8/16 bit pixel values in row-major order (refer to Figure 7 of the User Manual for pixels position and order). The byte order is little-endian for the 16 bit images.
        % The header is composed by a signature of 8 byte (0x4d5044ff03000001, starting with 4d on byte 0), and a metadata section of 1024 byte described in function HermesSaveImgDisk().
        % Hermes file can be read using the provided ImageJ/Fiji plugin.
        %
        % Parameters:
        %     filename: Full path of the output file.
        %     mode: File format of the output images
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP No images were acquired or the selected range of images is not valid
        %     UNABLE_CREATE_FILE Unable to create the output file

        % HermesSaveFlimDisk(Hermes_H Hermes, char* filename, OutFileFormat mode);
        obj.CheckEnumArg('mode', mode, obj.enuminfo.OutFileFormat);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSaveFlimDisk', obj.c_handle, filename, mode);
        % filename
        obj.checkError(ec)
    end
    function [Img, header] = ReadHermesFileFormatImage(obj, filename, ImgIdx, counter)
        % ReadHermesFileFormatImage - Read an integer (8 - 16 bit) Hermes image from file.
        %
        % ReadHermesFileFormatImage(ImgIdx, counter, Img, header)
        %
        % Read the image at the ImgIdx position and for desired counter in the given Hermes file from the hard disk.
        %
        % Parameters:
        %     filename: Full path of the output file.
        %     ImgIdx:    Image index in the file. Accepted values: 1 ... 65534
        %     counter: Desired counter. Accepted values: 1 ... 3
        %     Img: Pointer to the output image array. The size of the array must be at least 2 kiB.
        %     header: Array in which the header of Hermes file is saved.
        % Error codes:
        %     UNABLE_READ_FILE Unable to read the input file. Is it a Hermes file?
        %     OUT_OF_BOUND The desired counter or image exceeds the file size.
        %     NOT_EN_MEMORY Not enough memory to store the data contained in the file
        %     NULL_POINTER The provided provided handle or pointers point to an empty memory location.

        % HermesReadHermesFileFormatImage(char* filename, uint32_t ImgIdx, uint16_t counter, uint16_t* Img, char header[1024]);
        [ec, ~, Img, header] = calllib(obj.lib_alias, 'HermesReadHermesFileFormatImage', filename, ImgIdx, counter, zeros(32,64,'uint16'), ones(1,1024,'int8'));
        % filename, Img
        obj.checkError(ec)
    end
    function Img = AverageImg(obj, counter)
        % AverageImg - Calculate the average image.
        %
        % AverageImg(Img, counter)
        %
        % Once a set of images have been acquired by HermesSnapAcquire(), an image which contains for each pixel the average value over all the acquired images is
        % calculated. This is stored in the Img array.
        %
        % Parameters:
        %     counter: Desired counter. Accepted values: 1..3
        % Returns:
        %     The average image.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP No images were acquired

        % HermesAverageImg(Hermes_H Hermes, double* Img, uint16_t counter);
        data = zeros(obj.row_size*obj.num_rows, 1);
        [ec, ~, data] = calllib(obj.lib_alias, 'HermesAverageImg', obj.c_handle, data, counter);
        % Img
        obj.checkError(ec)

        frames = obj.BufferToFrames(data, obj.num_pixels , 1); %expect data of just one counter!
        Img = frames(:,:,1,1);
        
    end
    function Img = StDevImg(obj, counter)
        % StDevImg - Calculate the standard deviation image.
        %
        % StDevImg(Img, counter)
        %
        % Once a set of images have been acquired by HermesSnapAcquire(), an image which contains for each pixel the standard deviation over all the acquired images is
        % calculated.
        %
        % Parameters:
        %     counter: Desired counter. Accepted values: 1..3
        % Returns:
        %     The standard deviation image.
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP No images were acquired
        
        % HermesStDevImg(Hermes_H Hermes, double* Img, uint16_t counter);
        data = zeros(obj.row_size*obj.num_rows, 1);
        [ec, ~, data] = calllib(obj.lib_alias, 'HermesStDevImg', obj.c_handle, data, counter);
        % Img
        obj.checkError(ec)

        frames = obj.BufferToFrames(data, obj.num_pixels , 1); %expect data of just one counter!
        Img = frames(:,:,1,1);
    end
    function SetCorrelationMode(obj, CM, NCorrChannels, s)
        % SetCorrelationMode - Enable the correlation mode.
        %
        % SetCorrelationMode(CM, NCorrChannels, s)
        %
        % This function must be called before invoking HermesCorrelationImg(). When this function is called, the memory required to save the
        % new data is allocated in the heap and the previously stored data are cancelled. The deallocation of this memory is automatically performed
        % when the Hermes_destr() function is called or by setting the F equal to Disabled.
        %
        % Parameters:
        %     CM: Selected autocorrelation algorithm
        %     NCorrChannels: Number of global lag channels. When the linear correlation algorithm is selected, the first NChannel lags are calculated, where NChannel
        %     s: Enable or Disable the correlation mode
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     OUT_OF_BOUND NCorrChannels must be greater than zero for the Multi-tau algorithm and greater than 2 for the Linear one
        %     NOT_EN_MEMORY There is not enough memory to enable the correlation mode
        

        % HermesSetCorrelationMode(Hermes_H Hermes, CorrelationMode CM, int NCorrChannels, State s);
        obj.CheckEnumArg('CM', CM, obj.enuminfo.CorrelationMode);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSetCorrelationMode', obj.c_handle, CM, NCorrChannels, s);
        obj.checkError(ec)
    end
    function CorrelationImg(obj, counter)
        % CorrelationImg - Calculate the autocorrelation function.
        %
        % CorrelationImg(counter)
        %
        % The autocorrelation function is estimated for each pixel. This function requires that a set of
        % images have been previously acquired by HermesSnapAcquire() and that the correlation mode is set to Enabled. Depending on the selected algorithm and the
        % total number of collected images, this function can take several tens of seconds.
        %
        % Parameters:
        %     counter: Desired counter. Accepted values: 1..3
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP No images were acquired or the correlation mode was not enabled
        %     NOT_EN_MEMORY Not enough memory to calculate the correlation function
        %     INVALID_NIMG_CORRELATION The required number of time lags of the correlation function can not be calculated from the available number of images

        % HermesCorrelationImg(Hermes_H Hermes, uint16_t counter);
        obj.CheckBooleanArg('s', s);
        [ec, ~] = calllib(obj.lib_alias, 'HermesCorrelationImg', obj.c_handle, counter);
        obj.checkError(ec)
    end
    function [] = SaveCorrelationImg(obj, filename)
        % SaveCorrelationImg - Save the autocorrelation functions on the hard disk.
        %
        % SaveCorrelationImg(filename)
        %
        % This function requires that HermesSetCorrelationMode() and HermesCorrelationImg() have been previously called. The autocorrelation data are stored in a .hrmc binary
        % file.
        % The hrmc binary file is organized as follows:
        %
        % Byte offset               | Type           |Number of bytes | Description                                            |
        % --------------------------|----------------|----------------|-------------------------------------------------------|
        % 0                         | int            |4               | Number of lag-times (NLag)                            |
        % 4                         | int            |4               | Number of pixels. This value must be 1024 (NPix)        |
        % 8                         | int            |4               | Selected algorithm: 0 Linear, 1 Multi-tau                |
        % 12                        | double         |8 * NLag        | Autocorrelation values of the first pixel                |
        % 12 + 8 * NLag             | double         |8 * NLag        | Autocorrelation values of the second pixel            |
        % ...                       | double         |8 * NLag        | Autocorrelation values of the  N<SUP>th</SUP> pixel    |
        % 12 + 8 * (NPix-1) * NLag  | double         |8 * NLag        | Autocorrelation values of the last pixel                |
        % 12 + 8 * NPix * NLag      | double         |8 * NLag        | Lag times                                                |
        %
        % A simple Matlab script can be used to read the data for further processing or visualization.
        %
        % ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~{.mat}
        % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % MPD .HRMC file reader
        % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %
        % function data = Read_HRMC(fname)
        % f = fopen(fname,'rb');
        % buf = fread(f,3,'int32');
        % data.NChannel = buf(1);
        % data.NPixel = buf(2);
        % data.IsMultiTau = (buf(3) == 1);
        %
        % data.CorrelationImage = reshape(fread(f,data.NPixel*data.NChannel,'float64'), ...
        %             data.NChannel,32,32);
        % data.CorrelationImage = permute(data.CorrelationImage,[2 3 1]);
        % data.t=fread(f, data.NChannel,'float64');
        % fclose(f);
        % ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        %
        % Parameters:
        %     filename: Full path of the output file
        % Error codes:
        %     NULL_POINTER The provided Hermes_H points to an empty memory location
        %     INVALID_OP The autocorrelation, which has been calculated, is not valid
        %     UNABLE_CREATE_FILE Unable to create the output file
        %
        % HermesSaveCorrelationImg(Hermes_H Hermes, char* filename);
        [ec, ~] = calllib(obj.lib_alias, 'HermesSaveCorrelationImg', filename);
        % filename
        obj.checkError(ec)
    end
    function ResetOverilluminationProtection(obj)
        % ResetOverilluminationProtection - Reset the internal overillumination protection circuit.
        %
        % ResetOverilluminationProtection()
        %
        % This function resets the Hermes internal protection triggered by excessive illumination. It is mandatory to check that the overillumination condition is removed
        % before calling this function. In any case, in order to avoid damage to the Hermes, up to 3 reset cycles are allowed before having to disconnect the camera from the power supply.
        %
        % Parameters:
        %     None.
        % Error codes:
        %     TOO_MUCH_LIGHT The protection was reset, but the illumination is still too much!
        %     PERSISTING_TOO_MUCH_LIGHT The number of allowed reset cycles has been exceeded!


        % HermesResetOverilluminationProtection(Hermes_H Hermes);
        [ec, ~] = calllib(obj.lib_alias, 'HermesResetOverilluminationProtection', obj.c_handle);
        obj.checkError(ec)
    end


end


methods (Static, Access=public)  %%%% Hermes-SPECIFIC AUXILIARY %%%%
    function frames = BufferToFrames(data, num_pixels, num_counters)
        % BufferToFrames - Converts a plain flat Hermes data array such as data from ContAcqToMemoryGetBuffer()
        % to a more structured data set containing multiple frames
        %
        % frames = BufferToFrames(data, num_pixels, num_counters)
        %
        % Parameters
        %     data: 1-d array containing Hermes pixel data such as data returned by SnapGetImageBuffer(), ContAcqToMemoryGetBuffer()
        %     num_pixels: number of active pixels
        %     num_counters: number of active counters
        % Returns:
        %     Data reshaped as frames.
        
        if mod(numel(data), num_pixels * num_counters) ~= 0
            error('Hermes:invalidData', 'The number of elements of data must be a multiple of (num_pixels * num_counters)');
        end

        row_size = 32;
        
        num_full_rows = ceil(num_pixels/row_size);  % ceil division

        % reshape data in 'flat' blocks each containing all pixels of a frame
        all_flat_frames = reshape(data, num_pixels, numel(data) / num_pixels);
        
        % number of frames per counter
        num_frames = size(all_flat_frames, 2) / num_counters; % number of frames per counter
        
        if mod(num_pixels, row_size) == 0
            frames = reshape(all_flat_frames, row_size, num_full_rows, num_counters, num_frames);
        else
            % pad frames so that the number of pixels is a multiple of 32 pixels (i.e. one row)
            padded_frame_size = row_size * num_full_rows; % size of a padded flat frame (in whole rows)
            all_flat_frames_padded = zeros(padded_frame_size, numel(data) / num_pixels);
            all_flat_frames_padded(1:num_pixels, :) = all_flat_frames;

            % reshape data to return an array of frames
            frames = reshape(all_flat_frames_padded, row_size, num_full_rows, num_counters, num_frames);
        end
        
        
        % change indexing order:
        % from (row, column, counter, frame) to (row, column, frame, counter)
        frames = permute(frames, [1,2,4,3]);
    end

    function [frames, header, file_signature] = ReadHermesDataFile(path)
        % ReadHermesDataFile - reads .hrm acquisition files
        %
        % [frames, header, file_signature] = ReadHermesDataFile(path)
        %
        % Before reading the data, the function checks if the input is a valid Hermes file.
        % To disable the check, request the file_signature, as third return parameter.
        %
        % Parameters
        %     path: path to the .hrm data file
        % Returns:
        %     frames data, file header and file signature

        fid = fopen(path,'rb');

        file_signature = fread(fid, 8, '*char')';
        if nargout < 3 % check if the signature is the correct one
            hermes_id = [77 80 68 255 4 0 0 0];
            if any(file_signature ~= hermes_id)
                error('Hermes:invalidFile', 'The specified file %s is not a valid Hermes one.', path);
            end
        end

        header.camera_id = fread(fid, 10, '*char')';
        header.SN = fread(fid, 32, '*char')';
        header.FW_VER = fread(fid, 1, 'uint16')/100;
        header.custom_ver = char(fread(fid, 1, '*uint8')+ uint8('A')); % encoding 0 = A, 1= B etc.
        header.date_time = fread(fid, 20, '*char')';
        fread(fid, 35, '*char');

        header.N_rows = fread(fid, 1, 'uint8');
        header.N_cols = fread(fid, 1, 'uint8');
        header.bit_x_pix = fread(fid, 1, 'uint8');
        header.N_counters = fread(fid, 1, 'uint8');
        header.HwIntTime = fread(fid, 1, 'uint16')*10e-9;
        header.SummedFrames = fread(fid, 1, 'uint16');
        header.DeadTimeCorrectionON = fread(fid, 1, 'uint8')~= 0;
        header.GateDuty_C1 = fread(fid, 1, 'uint8');
        header.HoldOff = fread(fid, 1, 'uint16')*1e-9;
        header.BKGsubON = fread(fid, 1, 'uint8')~= 0;
        header.C1_2_signed = fread(fid, 1, 'uint8')~= 0;
        header.N_frames = fread(fid, 1, 'uint32');
        header.ImgAveraged = fread(fid, 1, 'uint8')~= 0;
        header.Caveraged = fread(fid, 1, 'uint8');
        header.N_ave = fread(fid, 1, 'uint16');
        header.GateDuty_C2 = fread(fid, 1, 'uint8');
        header.GateDuty_C3 = fread(fid, 1, 'uint8');
        header.Frames_x_syncIn = fread(fid, 1, 'uint16');
        header.N_pix = fread(fid, 1, 'uint16');
        fread(fid, 72, 'uint8');

        header.FLIM.ON = fread(fid, 1, 'uint8')~= 0;
        header.FLIM.shift_pct = fread(fid, 1, 'uint16');
        header.FLIM.steps = fread(fid, 1, 'uint16');
        header.FLIM.frameLen = fread(fid, 1, 'uint32')*10e-9;
        header.FLIM.binWidth = fread(fid, 1, 'uint16')*1e-15;
        fread(fid, 9, 'uint8');

        header.MultiGate.mode = fread(fid, 1, 'uint8');
        header.MultiGate.start_pos = fread(fid, 1, 'int16');
        header.MultiGate.widthC1 = fread(fid, 1, 'uint8');
        header.MultiGate.widthC2 = fread(fid, 1, 'uint8');
        header.MultiGate.widthC3 = fread(fid, 1, 'uint8');
        header.MultiGate.gapC1_2 = fread(fid, 1, 'uint16');
        header.MultiGate.gapC2_3 = fread(fid, 1, 'uint16');
        header.MultiGate.binWidth = fread(fid, 1, 'uint16')*1e-15;

        header.CoarseGate.C1.ON = fread(fid, 1, 'uint8')~= 0;
        header.CoarseGate.C1.startPos = fread(fid, 1, 'uint16')*10e-9;
        header.CoarseGate.C1.stopPos = fread(fid, 1, 'uint16')*10e-9;
        header.CoarseGate.C2.ON = fread(fid, 1, 'uint8')~= 0;
        header.CoarseGate.C2.startPos = fread(fid, 1, 'uint16')*10e-9;
        header.CoarseGate.C2.stopPos = fread(fid, 1, 'uint16')*10e-9;
        header.CoarseGate.C3.ON = fread(fid, 1, 'uint8')~= 0;
        header.CoarseGate.C3.startPos = fread(fid, 1, 'uint16')*10e-9;
        header.CoarseGate.C3.stopPos = fread(fid, 1, 'uint16')*10e-9;
        fread(fid, 53, 'uint8');

        header.PDE.ON = fread(fid, 1, 'uint8')~= 0;
        header.PDE.startWave = fread(fid, 1, 'uint16')*1e-9;
        header.PDE.stopWave = fread(fid, 1, 'uint16')*1e-9;
        header.PDE.step = fread(fid, 1, 'uint16')*1e-9;
        fread(fid, 717, 'uint8');


        if header.bit_x_pix == 16
            ty='*uint16';
        else
            ty='*uint8';
        end
        data_size = header.N_cols * header.N_rows * header.N_frames * header.N_counters;

        try
            data=fread(fid, data_size, ty);
        catch err
            fclose(fid);
            rethrow(err)
        end
        fclose(fid);

        num_pixels = header.N_pix;
        num_counters = header.N_counters;
        frames = Hermes.BufferToFrames(data, num_pixels, num_counters);
    end
    
end 

end % class