% --- EXAMPLES -------------------------------------------------------------------------------------------------

classdef HermesExamples < handle

methods (Static, Access=public)
    function Live()
        close('all')
        colormap('gray')

        h = Hermes('Normal');
        num_counters = 2;
        h.SetCameraPar(100, 1, 1000, num_counters, false, false, false);
        h.ApplySettings();
        h.LiveSetModeON();
        clf

        for k=1:1000
            live_frames = h.LiveGetImg();
            counter1_frame = live_frames(:,:,1);
            imagesc(counter1_frame);
            axis image
            title(sprintf('frame %d', k));
            drawnow;
            pause(0.010)
        end
        h.LiveSetModeOFF();
        clear h;
    end

    function LiveSubArray()
        close('all')
        colormap('gray')

        h = Hermes('Normal');
        num_rows = 8;
        num_pixels = 32 * num_rows;
        h.SetCameraParSubArray(1000, 1, 300, false, num_pixels);
        h.ApplySettings();
        h.LiveSetModeON();
        clf

        for k=1:1000
            live_frames = h.LiveGetImg();
            counter1_frame = live_frames(:,:,1);
            imagesc(counter1_frame);
            axis image
            title(sprintf('frame %d', k));
            drawnow;
            pause(0.010)
        end
        h.LiveSetModeOFF();
        clear h;
    end
    
    function BackgroundSubtraction()
        close all;
        h = Hermes('Normal');
        h.SetCameraPar(4096, 1000, 100, 1, false, false, false);
        h.ApplySettings();

        fprintf('Close the camera shutter and press ENTER...\n');
        pause();

        h.SnapPrepare();
        h.SnapAcquire();
        BackgroundImg = h.AverageImg(1);
        

        fprintf('Open the camera shutter and press ENTER...\n');
        pause();
        h.SetBackgroundSubtraction(false);
        h.ApplySettings();
        h.SnapPrepare();
        h.SnapAcquire();
        AvgImg_sub_off = h.AverageImg(1);

        h.SetBackgroundImg(BackgroundImg);
        h.SetBackgroundSubtraction(true);
        h.ApplySettings();
        h.SnapPrepare();
        h.SnapAcquire();
        AvgImg_sub_on = h.AverageImg(1);

        clf
        subplot(3,1,1)
        imagesc(BackgroundImg);
        axis image
        title('Background acquisition')

        subplot(3,1,2)
        imagesc(AvgImg_sub_off);
        axis image
        title('Image acquisition - without background subtraction')

        subplot(3,1,3)
        imagesc(AvgImg_sub_on);
        axis image
        title('Image acquisition - with background subtraction')

        clear h;
    end

    function Snap()
        close all;
        h = Hermes('Normal');

        num_frames = 100;
        num_counters = 1;
        h.SetCameraPar(100, num_frames, 300, num_counters, false, false, false);
        h.ApplySettings();
        h.SnapPrepare();
        h.SnapAcquire();
        frames = h.SnapGetImageBuffer();
        clf
        % plot first 9 frames
        for k=1:9
            subplot(3,3,k)
            imagesc(frames(:,:,k,1)); % all rows, all columns, k-th frame, counter 1
            axis image
        end

        clear h
    end

    function File()
        close all;
        h = Hermes('Normal');
        h.SetCameraPar(100, 300, 300, 1, false, false, false);
        h.ApplySettings();
        h.SnapPrepare();
        h.SnapAcquire();
        h.SaveImgDisk(1, 300, 'full_acquisition', h.OutFileFormat.HERMES_FILEFORMAT);
        counter = 1;
        is_double = false;
        h.SaveAveragedImgDisk(counter, 'avg_acquisition', h.OutFileFormat.HERMES_FILEFORMAT, is_double);
        clear h
        
        % read the file and extract data from the full acquisition
        [frames, file_header] = Hermes.ReadHermesDataFile('full_acquisition.hrm');

        % plot first 9 frames
        for k=1:9
            subplot(3,3,k)
            imagesc(frames(:,:,k,1)); % all rows, all columns, k-th frame, counter 1
            axis image
        end
        
    end

    function ContinuousFile()
        close all;
        h = Hermes('Normal');
        h.SetCameraPar(100, 300, 300, 1, false, false, false);
        h.ApplySettings();
        h.ContAcqToFileStart('continuous_acquisition');
        req_bytes = 1e3*2048*2; % 1k frames
        tot_bytes = 0;
        hwb = waitbar(0, 'Downloading Frames...');
        while tot_bytes < req_bytes
            tot_bytes = tot_bytes + h.ContAcqToFileGetMemory();
            waitbar(tot_bytes/req_bytes, hwb);
        end
        close(hwb)
        h.ContAcqToFileStop();
        clear h
    end

    function ContinuousInMemory()
        close all;
        h = Hermes('Normal');
        num_counters = 2; % use two counters in this example
        h.SetCameraPar(100, 1, 100, num_counters, false, false, false);
        h.ApplySettings();
        h.ContAcqToMemoryStart();
      
        data = [];
        total_num_values = 100*1000;
        while length(data) < total_num_values
            data = vertcat(data, h.ContAcqToMemoryGetBuffer()); % concatenate data from subsequent calls
            progress = 100. * numel(data) / total_num_values;
            if total_num_values > 0
                fprintf('Progress = %.1f%%\n', progress);
            end
        end
        h.ContAcqToMemoryStop();
        
        % trim end of data that may contain an incomplete frame 
        % or not the same number of frames per counter
        data = data(1:h.num_counters * h.num_pixels * floor(numel(data) / (h.num_counters * h.num_pixels)));
        
        frames = Hermes.BufferToFrames(data, h.num_pixels , h.num_counters);
        
        % plot first 4 frames of counter 1
        figure('Name', 'Frames from Counter 1');
        for k=1:4
            subplot(2,2,k)
            imagesc(frames(:,:,k,1)); % all rows, all columns, k-th frame, counter 1
            axis image
        end
        
        % plot first 4 frames of counter 2
        figure('Name', 'Frames from Counter 2');
        for k=1:4
            subplot(2,2,k)
            imagesc(frames(:,:,k,2)); % all rows, all columns, k-th frame, counter 2
            axis image
        end
        
    end
    
    function Flim()
        close('all')

        h = Hermes('Normal');

        FLIM_steps = 1000;
        FLIM_shift = 1;
        FLIM_start = -500;
        Length = 10;

        h.SetFlimState(true);
        flim_frame_time = h.SetFlimPar(FLIM_steps, FLIM_shift, FLIM_start, Length)

        h.SetTriggerOutState('None');
        h.ApplySettings();

        h.ContAcqToMemoryStart();


        fprintf('Downloading frames');

        data = [];
        total_num_values = FLIM_steps * 2048
        while length(data) < total_num_values
            data = vertcat(data, h.ContAcqToMemoryGetBuffer()); % concatenate data from subsequent calls
            progress = 100. * numel(data) / total_num_values;
            if total_num_values > 0
                fprintf('Progress = %.1f%%\n', progress);
            end
        end
        h.ContAcqToMemoryStop();


        all_frames = Hermes.BufferToFrames(data, h.num_pixels, h.num_counters);
        frames = all_frames(:,:,1:FLIM_steps,1); % from counter 1 only

        step_counts = zeros(FLIM_steps,1);

        for k = 1:1:FLIM_steps
            frame = frames(:,:,k);
            step_counts(k) = sum(frame(:));
        end

        subplot(2,1,1);
        imagesc(mean(frames, 3));
        title('Mean Frame');
        axis image;

        subplot(2,1,2);
        flim_period = 20;  % nanoseconds;
        plot_time = (1:1:FLIM_steps) * (flim_period / FLIM_steps);
        plot(plot_time, step_counts);
        title('FLIM mode - frame counts per step');
        xlabel('Time [ns]')
    end

end

end

